#!/usr/bin/env python

import nsysstats

class AsyncMemcpyPageable(nsysstats.Report):

    ROW_LIMIT = 50

    usage = f"""{{SCRIPT}}[:rows=<limit>] -- AsyncMemcpyPageable

    <limit> - Maximum number of rows returned by the query. If not given,
        only the top {ROW_LIMIT} results are displayed.

    Output:
        Duration : Duration of memcpy on GPU in nanoseconds
        Start Time : Start time of memcpy on GPU in nanoseconds
        Source Kind : Memcpy source memory kind
        Destination Kind : Memcpy destination memory kind
        Bytes : Number of bytes transferred
        PID : Process identifier
        Device ID : GPU device identifier
        Context ID : Context identifier
        Stream ID : Stream identifier
        API Name : Name of runtime API function

    This rule identifies asynchronous memory transfers that
    end up becoming synchronous if the memory is pageable.
"""

    query_format_columns = """
    SELECT
        duration AS "Duration (ns)",
        start AS "Start Time (ns)",
        srcKind AS "Source Kind",
        dstKind AS "Destination Kind",
        bytes AS "Bytes",
        pid AS "PID",
        deviceId AS "Device ID",
        contextId AS "Context ID",
        streamId AS "Stream ID",
        apiName AS "API Name"
        {HIDDEN_COL}
    FROM
        ({MEMCPY_FINAL})
"""

# This query returns all asynchronous memcpys using PAGEABLE memory.
    query_async_memcpy_pageable = """
WITH
    {MEM_KIND_STRS_CTE}
    runtime AS (
        SELECT
            correlationId,
            nameId
        FROM
            CUPTI_ACTIVITY_KIND_RUNTIME
    ),
    async AS (
        SELECT
            id,
            value
        FROM
            StringIds
        WHERE
            value LIKE '%cpy%async%'
    ),
    memcpy AS (
        SELECT
            *,
            ((globalPid >> 24) & 0x00FFFFFF) AS pid,
            globalPid
        FROM
            CUPTI_ACTIVITY_KIND_MEMCPY
        WHERE
            srcKind = 0 OR dstKind = 0
    )
    SELECT
        end - start AS duration,
        start,
        msrck.name AS srcKind,
        mdstk.name AS dstKind,
        copyKind,
        bytes,
        pid,
        globalPid,
        memcpy.deviceId,
        contextId,
        streamId,
        async.value AS apiName
    FROM
        memcpy
    JOIN
        async
    ON
        async.id = runtime.nameId
    JOIN
        runtime
    ON
        runtime.correlationId = memcpy.correlationId
    LEFT JOIN
        MemKindStrs AS msrck
    ON
        srcKind = msrck.id
    LEFT JOIN
        MemKindStrs AS mdstk
    ON
        dstKind = mdstk.id
    ORDER BY
        duration DESC
    LIMIT {ROW_LIMIT}
"""

    table_checks = {
        'CUPTI_ACTIVITY_KIND_RUNTIME':
            'There were no problems detected because the report does not'
            ' contain CUDA trace data.',
        'CUPTI_ACTIVITY_KIND_MEMCPY':
            'There were no problems detected because the report does not'
            ' contain CUDA trace data.',
    }

    def setup(self):
        err = super().setup()
        if err != None:
            return err

        row_limit = self.ROW_LIMIT
        hidden_col = ''
        for arg in self.args:
            s = arg.split('=', 1)
            if s[0] == 'rows' and s[1].isdigit():
                row_limit = s[1]
            # This flag is for internal use only.
            elif s[0] == 'GUI':
                hidden_col = ', globalPid AS "_Global PID", copyKind AS "_Copy Kind"'
            else:
                exit(self.EXIT_INVALID_ARG)

        qamp = self.query_async_memcpy_pageable.format(
            MEM_KIND_STRS_CTE = self.MEM_KIND_STRS_CTE,
            ROW_LIMIT = row_limit)

        self.query = self.query_format_columns.format(
            MEMCPY_FINAL = qamp,
            HIDDEN_COL = hidden_col)

if __name__ == "__main__":
    AsyncMemcpyPageable.Main()
